SystemOrganization addCategory: #'Zinc-Character-Encoding-Core'!

Object subclass: #ZnBase64Encoder
	instanceVariableNames: 'alphabet inverse lineLength lineEnd'
	classVariableNames: 'DefaultAlphabet DefaultInverse'
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnBase64Encoder commentStamp: '<historical>' prior: 0!
I am ZnBase64Encoder.

Base64 encoding is a technique to encode binary data as a string of characters that can be safely transported over various protocols. Basically, every 3 bytes are encoded using 4 characters from an alphabet of 64. Each encoded character represents 6 bits.

The most commonly used alphabet is 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'. One or two equal signs (= or ==) are used for padding.

The encoded data can optionally be broken into lines. Characters not part of the alphabet are considered as white space and are ignored.

My #encode: method works from ByteArray to String, while my #decode: method works from String to ByteArray.

Note that to encode a String as Base64, you first have to encode the characters as bytes using a character encoder.

See also http://en.wikipedia.org/wiki/Base64

Part of Zinc HTTP Components.!

!ZnBase64Encoder class methodsFor: 'class initialization' stamp: 'SvenVanCaekenberghe 12/16/2012 12:54'!
initialize
	DefaultAlphabet := String withAll: ($A to: $Z) , ($a to: $z) , ($0 to: $9) , #($+ $/).
	DefaultInverse := Array new: 128.
	0 to: 127 do: [ :each | 
		| offset |
		offset := DefaultAlphabet indexOf: each asCharacter ifAbsent: [ nil ].
		DefaultInverse at: each + 1 put: (offset ifNotNil: [ offset - 1 ]) ]! !

!ZnBase64Encoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/16/2012 15:15'!
alphabet: string
	"Set the alphabet to use to string, containing 64 characters to represent 64 byte values.
	I automatically compute the inverse used for fast decoding."
	
	self assert: string size = 64 description: '64 characters are needed for a Base64 alphabet'.
	alphabet := string.
	inverse := Array new: 128.
	0 to: 127 do: [ :each | 
		| offset |
		offset := alphabet indexOf: each asCharacter ifAbsent: [ nil ].
		inverse at: each + 1 put: (offset ifNotNil: [ offset - 1 ]) ]! !

!ZnBase64Encoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/15/2012 14:06'!
breakLines
	"Configure me to break lines and insert newlines every 76 characters while encoding"

	self breakLinesAt: 76! !

!ZnBase64Encoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/16/2012 15:13'!
breakLinesAt: length
	"Configure me to break lines at lenth, a multiple of 4, and insert newlines"

	self assert: (length \\ 4) isZero description: 'line length should be a multiple of 4'.
	lineLength := length.
	lineEnd ifNil: [ self lineEndConvention: #crlf ]! !

!ZnBase64Encoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/15/2012 14:07'!
decode: string
	"Decode a Base64 encoded string and return the resulting byte array"
	
	^ ByteArray
		new: string size
		streamContents: [ :byteStream | self decode: string readStream to: byteStream ]! !

!ZnBase64Encoder methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/15/2012 00:41'!
decode: char1 and: char2 and: char3 and: char4 to: stream
	| v1 v2 v3 v4 |
	v1 := self valueForCharacter: char1.
	v2 := self valueForCharacter: char2.
	v3 := self valueForCharacter: char3.
	v4 := self valueForCharacter: char4.
	stream nextPut: (v1 << 2 bitXor: (v2 ifNil: [ 0 ]) >> 4).
	char3 ~= $=
		ifTrue: [ 
			stream nextPut: ((v2 bitAnd: 2r1111) << 4 bitXor: (v3 ifNil: [ 0 ]) >> 2).
			char4 ~= $=
				ifTrue: [ stream nextPut: ((v3 bitAnd: 2r11) << 6 bitXor: v4) ] ]! !

!ZnBase64Encoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/16/2012 15:13'!
decode: stringStream to: byteStream
	[ stringStream atEnd ]
		whileFalse: [ 
			| char1 char2 char3 char4 |
			self skipWhiteSpace: stringStream.
			char1 := stringStream next.
			char2 := stringStream next.
			char3 := stringStream next.
			char4 := stringStream next.
			char1 isNil | char2 isNil | char3 isNil | char4 isNil
				ifTrue: [ ZnCharacterEncodingError signal: 'Illegal Base64 input' ].
			self
				decode: char1
				and: char2
				and: char3
				and: char4
				to: byteStream ]! !

!ZnBase64Encoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/15/2012 14:08'!
encode: byteArray
	"Encode byteArray using Base64 encoding and return the resulting string"
	
	^ String
		new: byteArray size
		streamContents: [ :stringStream | self encode: byteArray readStream to: stringStream ]! !

!ZnBase64Encoder methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/15/2012 00:42'!
encode: byte1 and: byte2 and: byte3 to: stream
	stream
		nextPut: (alphabet at: (byte1 >> 2) + 1);
		nextPut: (alphabet at: ((byte1 bitAnd: 2r11) << 4 bitXor: (byte2 ifNil: [ 0 ]) >> 4) + 1).
	byte2
		ifNil: [ stream nextPutAll: '==' ]
		ifNotNil: [ 
			stream nextPut: (alphabet at: ((byte2 bitAnd: 2r1111) << 2 bitXor: (byte3 ifNil: [ 0 ]) >> 6) + 1).
			byte3
				ifNil: [ stream nextPut: $= ]
				ifNotNil: [ stream nextPut: (alphabet at: (byte3 bitAnd: 2r111111) + 1) ] ]! !

!ZnBase64Encoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/15/2012 00:54'!
encode: byteStream to: stringStream
	| count |
	count := 0.
	[ byteStream atEnd ]
		whileFalse: [ 
			| byte1 byte2 byte3 |
			byte1 := byteStream next.
			byte2 := byteStream next.
			byte3 := byteStream next.
			self
				encode: byte1
				and: byte2
				and: byte3
				to: stringStream.
			count := count + 4.
			(lineLength notNil and: [ count = lineLength ])
				ifTrue: [ 
					stringStream nextPutAll: lineEnd.
					count := 0 ] ]! !

!ZnBase64Encoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/15/2012 13:48'!
initialize
	super initialize.
	alphabet := DefaultAlphabet.
	inverse := DefaultInverse ! !

!ZnBase64Encoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/15/2012 00:50'!
lineEndConvention: symbol
	"Set the end of line convention to be used.
	Either #cr, #lf or #crlf (the default)."
	
	self assert: (#(cr lf crlf) includes: symbol).
	lineEnd := String perform: symbol! !

!ZnBase64Encoder methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/16/2012 16:42'!
skipWhiteSpace: stream
	[ stream atEnd not and: [ (self valueForCharacter: stream peek) isNil ] ] 
		whileTrue: [ stream next ]! !

!ZnBase64Encoder methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/15/2012 00:31'!
valueForCharacter: char
	| code |
	code := char charCode.
	^ (code between: 0 and: 127)
		ifTrue: [ inverse at: code + 1 ]
		ifFalse: [ nil ]! !

Object subclass: #ZnBufferedReadStream
	instanceVariableNames: 'stream buffer position limit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnBufferedReadStream commentStamp: '<historical>' prior: 0!
I am STONBufferedReadStream.

I wrap another ReadStream and add efficient buffering for the typical access pattern of parsers: sending lots of #next, #peek and #atEnd messages.

By design I do not implement #position and #position: or anything based on that.

I can wrap both binary or character streams and act accordingly.

Part of Zinc HTTP Components.!

!ZnBufferedReadStream class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 11/30/2012 10:57'!
on: readStream
	^ self new
		on: readStream;
		yourself! !

!ZnBufferedReadStream class methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 12/2/2012 19:23'!
on: readStream do: block
	"Execute block with as argument a ZnBufferedReadStream on readStream.
	Return the value of block."

	^ block value: (self on: readStream)! !

!ZnBufferedReadStream methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/30/2012 10:39'!
atEnd
	^ position > limit and: [ stream atEnd ]
	! !

!ZnBufferedReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/30/2012 10:28'!
close
	stream close! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:18'!
collectionSpecies
	^ stream isBinary
		ifTrue: [ ByteArray ]
		ifFalse: [ String ]! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:15'!
defaultBufferSize
	^ 2 raisedToInteger: 16! !

!ZnBufferedReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/30/2012 20:15'!
initialize
	super initialize.
	position := 1.
	limit := 0! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 21:28'!
isBinary
	^ stream isBinary! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/1/2012 20:48'!
next
	"Return the next element and move over it"
	
	position > limit
		ifTrue: [ self nextBuffer ].
	^ position <= limit
		ifTrue: [ 
			| char |
			char := buffer at: position.
			position := position + 1.
			char ]
		ifFalse: [ nil ]! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:18'!
next: requestedCount 
	"Read requestedCount elements and return them as a collection.
	If less are available, a smaller collection will be returned."

	^ self 
		next: requestedCount 
		into: (self collectionSpecies new: requestedCount)! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:11'!
next: requestedCount into: collection
	"Read requestedCount elements into collection,
	returning a copy if less elements are available"
	
	^ self 
		next: requestedCount 
		into: collection 
		startingAt: 1   ! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:11'!
next: requestedCount into: collection startingAt: offset
	"Read requestedCount elements into collection starting at offset,
	returning a copy if less elements are available"
	
	| read |
	read := self 
		readInto: collection 
		startingAt: offset 
		count: requestedCount.
	^ read = requestedCount 
		ifTrue: [ collection ]
		ifFalse: [ collection copyFrom: 1 to: offset + read - 1 ]     ! !

!ZnBufferedReadStream methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/30/2012 11:15'!
nextBuffer
	limit := stream readInto: buffer startingAt: 1 count: buffer size.
	position := 1! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:11'!
nextInto: collection
	"Read the next elements of the receiver into collection,
	returning a copy if less elements are available"
	
	^ self
		next: collection size
		into: collection! !

!ZnBufferedReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/30/2012 20:15'!
on: readStream
	stream := readStream.
	self sizeBuffer: self defaultBufferSize! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/1/2012 20:48'!
peek
	"Return the next element but do not move over it"
	
	position > limit
		ifTrue: [ self nextBuffer ].
	^ position <= limit
		ifTrue: [ buffer at: position ]
		ifFalse: [ nil ]! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/1/2012 20:47'!
peekFor: object
	"Answer false and do not move over the next element if it is not equal to object, or if the receiver is at the end. 
	Answer true and move over the next element when it is equal to object."

	^ self peek = object
		ifTrue: [ 
			self next.
			true ]
		ifFalse: [ false ]! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 22:16'!
readInto: collection startingAt: offset count: requestedCount
	"Read requestedCount elements into collection starting at offset,
	returning the number of elements read, there could be less elements available."

	| read |
	read := 0.
	position <= limit
		ifTrue: [ 
			read := (limit - position + 1) min: requestedCount.
			collection replaceFrom: offset to: offset + read - 1 with: buffer startingAt: position.
			position := position + read.
			read = requestedCount ifTrue: [ ^ requestedCount ] ].
	^ read
		+
			(requestedCount - read < (buffer size / 2)
				ifTrue: [ 
					stream atEnd
						ifTrue: [ 0 ]
						ifFalse: [ 
							self
								nextBuffer;
								readInto: collection startingAt: offset + read count: requestedCount - read ] ]
				ifFalse: [ stream readInto: collection startingAt: offset + read count: requestedCount - read ])! !

!ZnBufferedReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/30/2012 20:19'!
sizeBuffer: size
	buffer := self collectionSpecies new: size! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 21:36'!
skip: count
	"Skip over count elements.
	This could be further optimzed."
	
	count timesRepeat: [ self next ]! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 21:33'!
upTo: value 
	"Read upto but not including value and return them as a collection.
	If value is not found, return the entire contents of the stream.
	This could be further optimzed."
	
	^ self collectionSpecies 
		streamContents: [ :writeStream | | element |
			[ self atEnd or: [ (element := self next) = value ] ] whileFalse: [ 
				writeStream nextPut: element ] ]! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/16/2012 16:31'!
upToEnd
	"Read elements until the stream is atEnd and return them as a collection."

	^ self collectionSpecies
		streamContents: [ :out | 
			[ self atEnd ] whileFalse: [ 
				position > limit
					ifTrue: [ self nextBuffer ].	
				out next: limit - position + 1 putAll: buffer startingAt: position.
				position := limit + 1 ] ]! !

!ZnBufferedReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 10:27'!
wrappedStream
	^ stream! !

Object subclass: #ZnBufferedWriteStream
	instanceVariableNames: 'stream buffer position'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnBufferedWriteStream commentStamp: '<historical>' prior: 0!
I am ZnBufferedWriteStream.
I wrap a write stream and add buffering.

Make sure to always send me #flush or #close when you're done,
otherwise the last buffer might not yet have been written.
My class side's #on:do: helps to ensure this.

I can wrap both binary or character streams and act accordingly.

Part of Zinc HTTP Components.!

!ZnBufferedWriteStream class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 12/7/2010 21:20'!
on: writeStream
	^ self basicNew
		on: writeStream;
		yourself! !

!ZnBufferedWriteStream class methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 12/2/2012 19:23'!
on: writeStream do: block
	"Execute block with as argument a ZnBufferedWriteStream on writeStream,
	making sure #flush is called at the end. Return the value of block."
	
	| bufferedWriteStream result |
	bufferedWriteStream := self on: writeStream.
	result := block value: bufferedWriteStream.
	bufferedWriteStream flush.
	^ result! !

!ZnBufferedWriteStream methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/8/2010 10:15'!
buffer
	buffer isNil 
		ifTrue: [ self sizeBuffer: self defaultBufferSize ].
	^ buffer! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/7/2010 23:42'!
bufferFreeSize
	^ self bufferSize - position! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2010 10:08'!
bufferSize
	^ buffer isNil 
		ifTrue: [ self defaultBufferSize ]
		ifFalse: [ buffer size ]! !

!ZnBufferedWriteStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/7/2010 21:22'!
close
	self flushBuffer.
	stream close! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 21:22'!
defaultBufferSize
	^ 2 raisedToInteger: 16! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/7/2010 21:23'!
flush
	self flushBuffer.
	stream flush! !

!ZnBufferedWriteStream methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2010 23:48'!
flushBuffer
	position = 0 ifTrue: [ ^ self ].
	position = self bufferSize
		ifTrue: [
			stream nextPutAll: buffer ]
		ifFalse: [
			(stream respondsTo: #next:putAll:startingAt:)
				ifTrue: [ stream next: position putAll: buffer startingAt: 1 ]
				ifFalse: [ stream nextPutAll: (buffer copyFrom: 1 to: position) ] ].
	position := 0! !

!ZnBufferedWriteStream methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2010 23:39'!
flushBufferIfFull
	position = self bufferSize 
		ifTrue: [ self flushBuffer ]
! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:01'!
next: count putAll: collection
	"Write count elements from collection"
	
	self 
		next: count 
		putAll: collection 
		startingAt: 1! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 21:20'!
next: count putAll: collection startingAt: offset
	"Write count elements from collection starting at offset."
	
	self flushBufferIfFull.
	count <= self bufferFreeSize
		ifTrue: [
			self buffer replaceFrom: position + 1 to: position + count with: collection startingAt: offset.
			position := position + count ]
		ifFalse: [
			self flushBuffer.
			count > (self bufferSize / 2)
				ifTrue: [ stream next: count putAll: collection startingAt: offset ]
				ifFalse: [ self next: count putAll: collection startingAt: offset ] ]
! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2010 10:16'!
nextPut: object
	self flushBufferIfFull.
	position := position + 1.
	self buffer at: position put: object! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 20:02'!
nextPutAll: collection
	"Write a collection"
	
	self 
		next: collection size 
		putAll: collection 
		startingAt: 1! !

!ZnBufferedWriteStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/7/2010 21:30'!
on: writeStream
	stream := writeStream.
	position := 0! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2012 22:21'!
print: object
	object printOn: self! !

!ZnBufferedWriteStream methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 12/7/2010 21:19'!
printOn: aStream
	aStream 
		nextPutAll: 'a '; 
		nextPutAll: self class name! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/7/2010 21:28'!
sizeBuffer: size
	buffer := (stream isBinary ifTrue: [ ByteArray ] ifFalse: [ String ]) new: size! !

!ZnBufferedWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/7/2010 21:20'!
space
	self nextPut: Character space! !

Object subclass: #ZnCharacterEncoder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnCharacterEncoder commentStamp: '<historical>' prior: 0!
I am ZnCharacterEncoder, I encode and decode Character objects to and from a binary stream.
I am an abstract class with following protocol:

#nextFromStream:
#nextPut:toStream:
#encodedByteCountFor:

The first two are compatible with TextConverter and subclasses.

I add some convenience methods:

#encodeString:
#decodeBytes:
#encodedByteCountForString:

Contrary to older encoders, I work strictly from strings to bytes and vice versa and I will throw errors instead of silently ignoring them.

Part of Zinc HTTP Components.!

ZnCharacterEncoder subclass: #ZnByteEncoder
	instanceVariableNames: 'identifier byteToUnicode unicodeToByte strict'
	classVariableNames: 'ByteTextConverters'
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnByteEncoder commentStamp: '<historical>' prior: 0!
I am ZnByteEncoder, a concrete subclass of ZnCharacterEncoder.
I handle single byte encodings where byte values 0 to 127 map to ASCII
and 128 to 255 are a permutation to Unicode characters.

I derive my mappings by parsing official unicode.org specifications.

Part of Zinc HTTP Components.!

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:34'!
cp1250Mapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WINDOWS/CP1250.TXT'"

	^ #(
	16r20AC nil 16r201A nil 16r201E 16r2026 16r2020 16r2021 
	nil 16r2030 16r0160 16r2039 16r015A 16r0164 16r017D 16r0179 
	nil 16r2018 16r2019 16r201C 16r201D 16r2022 16r2013 16r2014 
	nil 16r2122 16r0161 16r203A 16r015B 16r0165 16r017E 16r017A 
	16r00A0 16r02C7 16r02D8 16r0141 16r00A4 16r0104 16r00A6 16r00A7 
	16r00A8 16r00A9 16r015E 16r00AB 16r00AC 16r00AD 16r00AE 16r017B 
	16r00B0 16r00B1 16r02DB 16r0142 16r00B4 16r00B5 16r00B6 16r00B7 
	16r00B8 16r0105 16r015F 16r00BB 16r013D 16r02DD 16r013E 16r017C 
	16r0154 16r00C1 16r00C2 16r0102 16r00C4 16r0139 16r0106 16r00C7 
	16r010C 16r00C9 16r0118 16r00CB 16r011A 16r00CD 16r00CE 16r010E 
	16r0110 16r0143 16r0147 16r00D3 16r00D4 16r0150 16r00D6 16r00D7 
	16r0158 16r016E 16r00DA 16r0170 16r00DC 16r00DD 16r0162 16r00DF 
	16r0155 16r00E1 16r00E2 16r0103 16r00E4 16r013A 16r0107 16r00E7 
	16r010D 16r00E9 16r0119 16r00EB 16r011B 16r00ED 16r00EE 16r010F 
	16r0111 16r0144 16r0148 16r00F3 16r00F4 16r0151 16r00F6 16r00F7 
	16r0159 16r016F 16r00FA 16r0171 16r00FC 16r00FD 16r0163 16r02D9 )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:39'!
cp1252Mapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WINDOWS/CP1252.TXT'"

	^ #(
	16r20AC nil 16r201A 16r0192 16r201E 16r2026 16r2020 16r2021 
	16r02C6 16r2030 16r0160 16r2039 16r0152 nil 16r017D nil 
	nil 16r2018 16r2019 16r201C 16r201D 16r2022 16r2013 16r2014 
	16r02DC 16r2122 16r0161 16r203A 16r0153 nil 16r017E 16r0178 
	16r00A0 16r00A1 16r00A2 16r00A3 16r00A4 16r00A5 16r00A6 16r00A7 
	16r00A8 16r00A9 16r00AA 16r00AB 16r00AC 16r00AD 16r00AE 16r00AF 
	16r00B0 16r00B1 16r00B2 16r00B3 16r00B4 16r00B5 16r00B6 16r00B7 
	16r00B8 16r00B9 16r00BA 16r00BB 16r00BC 16r00BD 16r00BE 16r00BF 
	16r00C0 16r00C1 16r00C2 16r00C3 16r00C4 16r00C5 16r00C6 16r00C7 
	16r00C8 16r00C9 16r00CA 16r00CB 16r00CC 16r00CD 16r00CE 16r00CF 
	16r00D0 16r00D1 16r00D2 16r00D3 16r00D4 16r00D5 16r00D6 16r00D7 
	16r00D8 16r00D9 16r00DA 16r00DB 16r00DC 16r00DD 16r00DE 16r00DF 
	16r00E0 16r00E1 16r00E2 16r00E3 16r00E4 16r00E5 16r00E6 16r00E7 
	16r00E8 16r00E9 16r00EA 16r00EB 16r00EC 16r00ED 16r00EE 16r00EF 
	16r00F0 16r00F1 16r00F2 16r00F3 16r00F4 16r00F5 16r00F6 16r00F7 
	16r00F8 16r00F9 16r00FA 16r00FB 16r00FC 16r00FD 16r00FE 16r00FF )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:35'!
cp1253Mapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WINDOWS/CP1253.TXT'"

	^ #(
	16r20AC nil 16r201A 16r0192 16r201E 16r2026 16r2020 16r2021 
	nil 16r2030 nil 16r2039 nil nil nil nil 
	nil 16r2018 16r2019 16r201C 16r201D 16r2022 16r2013 16r2014 
	nil 16r2122 nil 16r203A nil nil nil nil 
	16r00A0 16r0385 16r0386 16r00A3 16r00A4 16r00A5 16r00A6 16r00A7 
	16r00A8 16r00A9 nil 16r00AB 16r00AC 16r00AD 16r00AE 16r2015 
	16r00B0 16r00B1 16r00B2 16r00B3 16r0384 16r00B5 16r00B6 16r00B7 
	16r0388 16r0389 16r038A 16r00BB 16r038C 16r00BD 16r038E 16r038F 
	16r0390 16r0391 16r0392 16r0393 16r0394 16r0395 16r0396 16r0397 
	16r0398 16r0399 16r039A 16r039B 16r039C 16r039D 16r039E 16r039F 
	16r03A0 16r03A1 nil 16r03A3 16r03A4 16r03A5 16r03A6 16r03A7 
	16r03A8 16r03A9 16r03AA 16r03AB 16r03AC 16r03AD 16r03AE 16r03AF 
	16r03B0 16r03B1 16r03B2 16r03B3 16r03B4 16r03B5 16r03B6 16r03B7 
	16r03B8 16r03B9 16r03BA 16r03BB 16r03BC 16r03BD 16r03BE 16r03BF 
	16r03C0 16r03C1 16r03C2 16r03C3 16r03C4 16r03C5 16r03C6 16r03C7 
	16r03C8 16r03C9 16r03CA 16r03CB 16r03CC 16r03CD 16r03CE nil )! !

!ZnByteEncoder class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/15/2012 20:02'!
generateByteToUnicodeSpec: url
	"Return the formatted source code for an array mapping 
	the top 128 byte to unicode values from a Unicode.org url"
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/ISO8859/8859-2.TXT'."

	| mapping |
	mapping := self parseUnicodeOrgSpec: url.  
	^ String streamContents: [ :stream |
		stream tab; << '"'; << 'self generateByteToUnicodeSpec: '; print: url; << '"'; cr; cr; tab; << '^ #('.
		128 to: 255 do: [ :each | | unicode |
			each \\ 8 = 0 ifTrue: [ stream cr; tab ].
			(unicode := mapping at: each ifAbsent: [ nil ]) isNil
				ifTrue: [ stream print: nil; space ]  
				ifFalse: [ stream << '16r' << (unicode printPaddedWith: $0 to: 4 base: 16); space ] ].
		stream nextPut: $); cr ]! !

!ZnByteEncoder class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 1/25/2011 13:02'!
handlesEncoding: string
	"Return true when my instances handle the encoding described by string"
	
	^ ByteTextConverters includesKey: string! !

!ZnByteEncoder class methodsFor: 'class initialization' stamp: 'SvenVanCaekenberghe 1/15/2013 17:03'!
initialize
	"Initialize and cache the converters that I know of.
	This method must be changed to make sure it runs 
	when loading in images where it is already present."
	
	self initializeByteTextConverters! !

!ZnByteEncoder class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 1/15/2013 17:01'!
initializeByteTextConverters
	"Initialize and cache convertors based on specifications in methods that were autogenerated."

	ByteTextConverters := Dictionary new.
	self mappingToIdentifiers
		keysAndValuesDo: [ :mapping :identifiers | 
			| tables |
			tables := self tablesFromSpec: (self perform: mapping).
			identifiers do: [ :each | ByteTextConverters at: each put: tables ] ]! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:37'!
iso885915Mapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/ISO8859/8859-15.TXT'"

	^ #(
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	16r00A0 16r00A1 16r00A2 16r00A3 16r20AC 16r00A5 16r0160 16r00A7 
	16r0161 16r00A9 16r00AA 16r00AB 16r00AC 16r00AD 16r00AE 16r00AF 
	16r00B0 16r00B1 16r00B2 16r00B3 16r017D 16r00B5 16r00B6 16r00B7 
	16r017E 16r00B9 16r00BA 16r00BB 16r0152 16r0153 16r0178 16r00BF 
	16r00C0 16r00C1 16r00C2 16r00C3 16r00C4 16r00C5 16r00C6 16r00C7 
	16r00C8 16r00C9 16r00CA 16r00CB 16r00CC 16r00CD 16r00CE 16r00CF 
	16r00D0 16r00D1 16r00D2 16r00D3 16r00D4 16r00D5 16r00D6 16r00D7 
	16r00D8 16r00D9 16r00DA 16r00DB 16r00DC 16r00DD 16r00DE 16r00DF 
	16r00E0 16r00E1 16r00E2 16r00E3 16r00E4 16r00E5 16r00E6 16r00E7 
	16r00E8 16r00E9 16r00EA 16r00EB 16r00EC 16r00ED 16r00EE 16r00EF 
	16r00F0 16r00F1 16r00F2 16r00F3 16r00F4 16r00F5 16r00F6 16r00F7 
	16r00F8 16r00F9 16r00FA 16r00FB 16r00FC 16r00FD 16r00FE 16r00FF )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:36'!
iso88591Mapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/ISO8859/8859-1.TXT'"

	^ #(
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	16r00A0 16r00A1 16r00A2 16r00A3 16r00A4 16r00A5 16r00A6 16r00A7 
	16r00A8 16r00A9 16r00AA 16r00AB 16r00AC 16r00AD 16r00AE 16r00AF 
	16r00B0 16r00B1 16r00B2 16r00B3 16r00B4 16r00B5 16r00B6 16r00B7 
	16r00B8 16r00B9 16r00BA 16r00BB 16r00BC 16r00BD 16r00BE 16r00BF 
	16r00C0 16r00C1 16r00C2 16r00C3 16r00C4 16r00C5 16r00C6 16r00C7 
	16r00C8 16r00C9 16r00CA 16r00CB 16r00CC 16r00CD 16r00CE 16r00CF 
	16r00D0 16r00D1 16r00D2 16r00D3 16r00D4 16r00D5 16r00D6 16r00D7 
	16r00D8 16r00D9 16r00DA 16r00DB 16r00DC 16r00DD 16r00DE 16r00DF 
	16r00E0 16r00E1 16r00E2 16r00E3 16r00E4 16r00E5 16r00E6 16r00E7 
	16r00E8 16r00E9 16r00EA 16r00EB 16r00EC 16r00ED 16r00EE 16r00EF 
	16r00F0 16r00F1 16r00F2 16r00F3 16r00F4 16r00F5 16r00F6 16r00F7 
	16r00F8 16r00F9 16r00FA 16r00FB 16r00FC 16r00FD 16r00FE 16r00FF )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:36'!
iso88592Mapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/ISO8859/8859-2.TXT'"

	^ #(
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	16r00A0 16r0104 16r02D8 16r0141 16r00A4 16r013D 16r015A 16r00A7 
	16r00A8 16r0160 16r015E 16r0164 16r0179 16r00AD 16r017D 16r017B 
	16r00B0 16r0105 16r02DB 16r0142 16r00B4 16r013E 16r015B 16r02C7 
	16r00B8 16r0161 16r015F 16r0165 16r017A 16r02DD 16r017E 16r017C 
	16r0154 16r00C1 16r00C2 16r0102 16r00C4 16r0139 16r0106 16r00C7 
	16r010C 16r00C9 16r0118 16r00CB 16r011A 16r00CD 16r00CE 16r010E 
	16r0110 16r0143 16r0147 16r00D3 16r00D4 16r0150 16r00D6 16r00D7 
	16r0158 16r016E 16r00DA 16r0170 16r00DC 16r00DD 16r0162 16r00DF 
	16r0155 16r00E1 16r00E2 16r0103 16r00E4 16r013A 16r0107 16r00E7 
	16r010D 16r00E9 16r0119 16r00EB 16r011B 16r00ED 16r00EE 16r010F 
	16r0111 16r0144 16r0148 16r00F3 16r00F4 16r0151 16r00F6 16r00F7 
	16r0159 16r016F 16r00FA 16r0171 16r00FC 16r00FD 16r0163 16r02D9 )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:39'!
iso88597Mapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/ISO8859/8859-7.TXT'"

	^ #(
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	nil nil nil nil nil nil nil nil 
	16r00A0 16r2018 16r2019 16r00A3 16r20AC 16r20AF 16r00A6 16r00A7 
	16r00A8 16r00A9 16r037A 16r00AB 16r00AC 16r00AD nil 16r2015 
	16r00B0 16r00B1 16r00B2 16r00B3 16r0384 16r0385 16r0386 16r00B7 
	16r0388 16r0389 16r038A 16r00BB 16r038C 16r00BD 16r038E 16r038F 
	16r0390 16r0391 16r0392 16r0393 16r0394 16r0395 16r0396 16r0397 
	16r0398 16r0399 16r039A 16r039B 16r039C 16r039D 16r039E 16r039F 
	16r03A0 16r03A1 nil 16r03A3 16r03A4 16r03A5 16r03A6 16r03A7 
	16r03A8 16r03A9 16r03AA 16r03AB 16r03AC 16r03AD 16r03AE 16r03AF 
	16r03B0 16r03B1 16r03B2 16r03B3 16r03B4 16r03B5 16r03B6 16r03B7 
	16r03B8 16r03B9 16r03BA 16r03BB 16r03BC 16r03BD 16r03BE 16r03BF 
	16r03C0 16r03C1 16r03C2 16r03C3 16r03C4 16r03C5 16r03C6 16r03C7 
	16r03C8 16r03C9 16r03CA 16r03CB 16r03CC 16r03CD 16r03CE nil )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:38'!
koi8rMapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/VENDORS/MISC/KOI8-R.TXT'"

	^ #(
	16r2500 16r2502 16r250C 16r2510 16r2514 16r2518 16r251C 16r2524 
	16r252C 16r2534 16r253C 16r2580 16r2584 16r2588 16r258C 16r2590 
	16r2591 16r2592 16r2593 16r2320 16r25A0 16r2219 16r221A 16r2248 
	16r2264 16r2265 16r00A0 16r2321 16r00B0 16r00B2 16r00B7 16r00F7 
	16r2550 16r2551 16r2552 16r0451 16r2553 16r2554 16r2555 16r2556 
	16r2557 16r2558 16r2559 16r255A 16r255B 16r255C 16r255D 16r255E 
	16r255F 16r2560 16r2561 16r0401 16r2562 16r2563 16r2564 16r2565 
	16r2566 16r2567 16r2568 16r2569 16r256A 16r256B 16r256C 16r00A9 
	16r044E 16r0430 16r0431 16r0446 16r0434 16r0435 16r0444 16r0433 
	16r0445 16r0438 16r0439 16r043A 16r043B 16r043C 16r043D 16r043E 
	16r043F 16r044F 16r0440 16r0441 16r0442 16r0443 16r0436 16r0432 
	16r044C 16r044B 16r0437 16r0448 16r044D 16r0449 16r0447 16r044A 
	16r042E 16r0410 16r0411 16r0426 16r0414 16r0415 16r0424 16r0413 
	16r0425 16r0418 16r0419 16r041A 16r041B 16r041C 16r041D 16r041E 
	16r041F 16r042F 16r0420 16r0421 16r0422 16r0423 16r0416 16r0412 
	16r042C 16r042B 16r0417 16r0428 16r042D 16r0429 16r0427 16r042A )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 19:38'!
macRomanMapping
	"self generateByteToUnicodeSpec: 'http://unicode.org/Public/MAPPINGS/VENDORS/APPLE/ROMAN.TXT'"

	^ #(
	16r00C4 16r00C5 16r00C7 16r00C9 16r00D1 16r00D6 16r00DC 16r00E1 
	16r00E0 16r00E2 16r00E4 16r00E3 16r00E5 16r00E7 16r00E9 16r00E8 
	16r00EA 16r00EB 16r00ED 16r00EC 16r00EE 16r00EF 16r00F1 16r00F3 
	16r00F2 16r00F4 16r00F6 16r00F5 16r00FA 16r00F9 16r00FB 16r00FC 
	16r2020 16r00B0 16r00A2 16r00A3 16r00A7 16r2022 16r00B6 16r00DF 
	16r00AE 16r00A9 16r2122 16r00B4 16r00A8 16r2260 16r00C6 16r00D8 
	16r221E 16r00B1 16r2264 16r2265 16r00A5 16r00B5 16r2202 16r2211 
	16r220F 16r03C0 16r222B 16r00AA 16r00BA 16r03A9 16r00E6 16r00F8 
	16r00BF 16r00A1 16r00AC 16r221A 16r0192 16r2248 16r2206 16r00AB 
	16r00BB 16r2026 16r00A0 16r00C0 16r00C3 16r00D5 16r0152 16r0153 
	16r2013 16r2014 16r201C 16r201D 16r2018 16r2019 16r00F7 16r25CA 
	16r00FF 16r0178 16r2044 16r20AC 16r2039 16r203A 16rFB01 16rFB02 
	16r2021 16r00B7 16r201A 16r201E 16r2030 16r00C2 16r00CA 16r00C1 
	16r00CB 16r00C8 16r00CD 16r00CE 16r00CF 16r00CC 16r00D3 16r00D4 
	16rF8FF 16r00D2 16r00DA 16r00DB 16r00D9 16r0131 16r02C6 16r02DC 
	16r00AF 16r02D8 16r02D9 16r02DA 16r00B8 16r02DD 16r02DB 16r02C7 )! !

!ZnByteEncoder class methodsFor: 'mappings' stamp: 'SvenVanCaekenberghe 12/15/2012 21:38'!
mappingToIdentifiers
	"Return a dictionay mapping from encoding specifications to a list of encoding names."
	
	^ Dictionary newFromPairs: #(
		#cp1250Mapping #('cp-1250') 
		#cp1252Mapping #('cp-1252')
		#cp1253Mapping #('cp1253')
		#iso885915Mapping #('iso-8859-15') 
		#iso88591Mapping #('iso-8859-1' 'latin1' 'latin-1') 
		#iso88592Mapping #('iso-8859-2' 'latin2' 'latin-2') 
		#iso88597Mapping #('iso-8859-7') 
		#koi8rMapping #('iso-8859-15')
		#macRomanMapping #('mac-roman' 'macintosh') )! !

!ZnByteEncoder class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 12/15/2012 21:58'!
newForEncoding: string
	"Return a new character encoder object for an encoding described by string.
	We use our precomputed ByteTextConverters tables."

	| tables |
	tables := ByteTextConverters at: string.
	^ self new
		identifier: string;
		byteToUnicode: tables first;
		unicodeToByte: tables second;
		yourself! !

!ZnByteEncoder class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 1/15/2013 14:54'!
parseUnicodeOrgSpec: url
	"Parse and return a mapping from byte to unicode values from url."
	"self parseUnicodeOrgSpec: 'http://unicode.org/Public/MAPPINGS/ISO8859/8859-2.TXT'."
	
	| mapping |
	mapping := Dictionary new: 256.
	url asZnUrl retrieveContents linesDo: [ :each |
		(each isEmpty or: [ each beginsWith: '#' ])
			ifFalse: [ | tokens hexReader |
				hexReader := [ :string | Integer readFrom: (string readStream skip: 2; yourself) base: 16 ].
				tokens := each findTokens: String tab. 
				(tokens last = '<control>' or: [ tokens last = '#UNDEFINED' ]) ifFalse: [
					mapping
						at: (hexReader value: tokens first)
						put: (hexReader value: tokens second) ] ] ].
	^ mapping ! !

!ZnByteEncoder class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/15/2012 21:31'!
tablesFromSpec: mapping
	"Initialize the mappings to and from Unicode based on the 128 element array mapping"

	| byteToUnicode unicodeToByte |
	byteToUnicode := Array new: 128.
	unicodeToByte := Dictionary new.	
	"Mind the offset because first 128 characters are not stored into byteToUnicodeSpec"	
	"Note that some entries are nil"
	mapping
		keysAndValuesDo: [ :index :unicode | 
			unicode ifNotNil: [ 
				byteToUnicode at: index put: (Character value: unicode).
				unicodeToByte at: unicode put: 127 + index ] ].
	^ Array with: byteToUnicode with: unicodeToByte! !

!ZnByteEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/17/2012 15:40'!
beLenient
	"Don't be strict, which is the default.
	This means that holes in the mapping are let to pass through."

	strict := false! !

!ZnByteEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 1/25/2011 12:31'!
byteToUnicode: map
	byteToUnicode := map! !

!ZnByteEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 1/25/2011 12:19'!
encodedByteCountFor: character
	^ 1! !

!ZnByteEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/15/2012 21:51'!
identifier: object
	identifier := object! !

!ZnByteEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/17/2012 14:46'!
initialize
	super initialize.
	strict := true! !

!ZnByteEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/17/2012 15:38'!
nextFromStream: stream
	"In non-strict mode, we let byte values for holes in our mapping pass through"

	| byteValue |
	^ (byteValue := stream next) < 128
		ifTrue: [ Character value: byteValue ]
		ifFalse: [ 
			(byteToUnicode at: byteValue - 127 ifAbsent: [ nil ])
				ifNotNil: [ :unicode | unicode ]
				ifNil: [ 
					strict
						ifTrue: [ self error: 'Character outside encoder range' ]
						ifFalse: [ Character value: byteValue ] ] ]! !

!ZnByteEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/17/2012 15:39'!
nextPut: character toStream: stream
	"In non-strict mode, we let characters for holes in our mapping table pass through"
	
	| code |
	(code := character codePoint) < 128
		ifTrue: [ stream nextPut: code ]
		ifFalse: [ 
			| byte |
			byte := unicodeToByte at: code ifAbsent: [ nil ].
			(byte isNil and: [ strict or: [ code > 255 ] ])
				ifTrue: [ self error: 'Character code outside encoder range' ].
			stream nextPut: (byte ifNil: [ code ]) ]! !

!ZnByteEncoder methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 1/15/2013 16:58'!
printOn: stream
	super printOn: stream.
	stream
		nextPut: $(;
		print: identifier.
	strict
		ifTrue: [ stream nextPutAll: ' strict' ].
	stream nextPut: $)! !

!ZnByteEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 1/25/2011 12:31'!
unicodeToByte: map
	unicodeToByte := map! !

!ZnCharacterEncoder class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 1/25/2011 11:07'!
handlesEncoding: string
	"Return true when my instances handle the encoding described by string"
	
	self subclassResponsibility ! !

!ZnCharacterEncoder class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 1/25/2011 13:27'!
newForEncoding: string
	"Return a new character encoder object for an encoding described by string.
	Search for a subclass that handles it and delegate (subclassResponsibility).
	We default to ZnNullEncoder."
	
	| concreteSubclass encoding |
	encoding := string asLowercase.
	concreteSubclass := self subclasses 
		detect: [ :each | each handlesEncoding: encoding ] 
		ifNone: [ ZnNullEncoder ].
	^ concreteSubclass newForEncoding: encoding! !

!ZnCharacterEncoder methodsFor: 'comparing' stamp: 'SvenVanCaekenberghe 12/14/2010 11:56'!
= anObject
	^ self class == anObject class! !

!ZnCharacterEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/17/2012 14:47'!
beLenient
	"Don't be strict, which is the default"! !

!ZnCharacterEncoder methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 12/16/2012 15:17'!
decodeBytes: bytes
	"Decode bytes and return the resulting string"
	
	| byteStream |
	byteStream := bytes readStream.
	^ String streamContents: [ :stream |
		[ byteStream atEnd ] whileFalse: [
			stream nextPut: (self nextFromStream: byteStream) ] ] ! !

!ZnCharacterEncoder methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 12/16/2012 15:17'!
encodeString: string
	"Encode string and return the resulting byte array"
	
	^ ByteArray streamContents: [ :stream |
		string do: [ :each | 
			self nextPut: each toStream: stream ] ] ! !

!ZnCharacterEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 11/29/2010 21:17'!
encodedByteCountFor: character
	"Return how many bytes are needed to encode character"
	
	self subclassResponsibility ! !

!ZnCharacterEncoder methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 12/16/2012 15:18'!
encodedByteCountForString: string
	"Return the exact number of bytes it would take to encode string as a byte array"

	^ string 
		inject: 0 
		into: [ :sum :each |
			sum + (self encodedByteCountFor: each) ]! !

!ZnCharacterEncoder methodsFor: 'error handling' stamp: 'SvenVanCaekenberghe 12/16/2012 15:21'!
error: message
	ZnCharacterEncodingError signal: message! !

!ZnCharacterEncoder methodsFor: 'comparing' stamp: 'SvenVanCaekenberghe 4/20/2011 12:19'!
hash
	^ self class name hash! !

!ZnCharacterEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 11/29/2010 21:16'!
nextFromStream: stream
	"Read and return the next character from stream"
	
	self subclassResponsibility ! !

!ZnCharacterEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 11/29/2010 21:16'!
nextPut: character toStream: stream
	"Write the encoding for character to stream"
	
	self subclassResponsibility ! !

ZnCharacterEncoder subclass: #ZnNullEncoder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnNullEncoder commentStamp: '<historical>' prior: 0!
I am ZnNullEncoder, a concrete subclass of ZnCharacterEncoder.
I perform no encoding or decoding at all for all characters with a code value below 256.
I can only be used for ASCII.

Note that in principle I could handle Latin1 (ISO-8859-1), although that is not completely correct. To get maximum efficiency, it remains an option.
	
Part of Zinc HTTP Components.!

!ZnNullEncoder class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/17/2012 13:24'!
handlesEncoding: string
	"Return true when my instances handle the encoding described by string.
	Note that in principle I could handle latin1 (iso-8859-1), although that is not 
	completely correct. To get maximum efficiency, it remains an option."
	
	^ #( 'ascii' ) includes: string! !

!ZnNullEncoder class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 1/25/2011 11:19'!
newForEncoding: string
	"No further parametrization needed"
	
	^ self new! !

!ZnNullEncoder methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 5/3/2012 11:32'!
decodeBytes: bytes
	^ bytes asString! !

!ZnNullEncoder methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 5/3/2012 11:32'!
encodeString: string
	^ string asByteArray! !

!ZnNullEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 11/29/2010 21:17'!
encodedByteCountFor: character
	^ 1! !

!ZnNullEncoder methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 5/3/2012 11:32'!
encodedByteCountForString: string
	^ string size! !

!ZnNullEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 1/4/2011 20:52'!
nextFromStream: stream
	^ Character codePoint: stream next! !

!ZnNullEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 1/25/2011 12:36'!
nextPut: character toStream: stream
	| code |
	(code := character charCode) < 256
		ifTrue: [ stream nextPut: code ]
		ifFalse: [ self error: 'Character code outside encoder range' ]! !

ZnCharacterEncoder subclass: #ZnUTF8Encoder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnUTF8Encoder commentStamp: 'SvenVanCaekenberghe 12/6/2010 12:55' prior: 0!
I am ZnUTF8Encoder, a concrete subclass of ZnCharacterEncoder.
I implement the variable length UTF-8 encoding and decoding of Unicode according to RFC 3629.

Part of Zinc HTTP Components.!

!ZnUTF8Encoder class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 1/25/2011 11:11'!
handlesEncoding: string
	"Return true when my instances handle the encoding described by string"
	
	^ #( 'utf-8' 'utf8' ) includes: string! !

!ZnUTF8Encoder class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 1/25/2011 11:19'!
newForEncoding: string
	"No further parametrization needed"
	
	^ self new! !

!ZnUTF8Encoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 11/29/2010 21:23'!
encodedByteCountFor: character
	| code |
	code := character charCode.
	code < 128 ifTrue: [ ^ 1 ].
	code < 2048 ifTrue: [ ^ 2 ].
	code < 65535 ifTrue: [ ^ 3 ].
	code <= self maximumUTF8Code ifTrue: [ ^ 4 ].
	self error: 'Unicode character code point outside range'
! !

!ZnUTF8Encoder methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/29/2010 21:19'!
maximumUTF8Code
	^ 16r10FFFF! !

!ZnUTF8Encoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/3/2012 14:25'!
nextFromStream: stream
	| code byte next |
	(byte := stream next) < 128
		ifTrue: [ ^ Character codePoint: byte ].
	(byte bitAnd: 2r11100000) == 2r11000000
		ifTrue: [ 
			code := byte bitAnd: 2r00011111.
			((next := stream next) bitAnd: 2r11000000) == 2r10000000
				ifTrue: [ code := (code bitShift: 6) + (next bitAnd: 2r00111111) ]
				ifFalse: [ ^ self error: 'illegal UTF-8 encoding' ].
			^ Character codePoint: code ].
	(byte bitAnd: 2r11110000) == 2r11100000
		ifTrue: [ 
			code := byte bitAnd: 2r00001111.
			2 timesRepeat: [ 
				((next := stream next) bitAnd: 2r11000000) == 2r10000000
					ifTrue: [ code := (code bitShift: 6) + (next bitAnd: 2r00111111) ]
					ifFalse: [ ^ self error: 'illegal UTF-8 encoding' ] ].
			^ Character codePoint: code ].
	(byte bitAnd: 2r11111000) == 2r11110000
		ifTrue: [ 
			code := byte bitAnd: 2r00000111.
			3 timesRepeat: [ 
				((next := stream next) bitAnd: 2r11000000) == 2r10000000
					ifTrue: [ code := (code bitShift: 6) + (next bitAnd: 2r00111111) ]
					ifFalse: [ ^ self error: 'illegal UTF-8 encoding' ] ].
			^ Character codePoint: code ].
	self error: 'illegal UTF-8 encoding'! !

!ZnUTF8Encoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/16/2012 15:20'!
nextPut: character toStream: stream
	| code |
	code := character codePoint.
	code < 128 ifTrue: [ 
		^ stream nextPut: code ].
	code < 2048 ifTrue: [ 
		^ stream 
			nextPut: (2r11000000 + (code bitShift: -6)); 
			nextPut: (2r10000000 + (code bitAnd: 2r111111)) ].
	code < 65535 ifTrue: [ 
		^ stream 
			nextPut: (2r11100000 + (code bitShift: -12));
			nextPut: (2r10000000 + ((code bitShift: -6) bitAnd: 2r111111)); 
			nextPut: (2r10000000 + (code bitAnd: 2r111111)) ].
	code <= self maximumUTF8Code ifTrue: [ 
		^ stream 
			nextPut: (2r11110000 + (code bitShift: -18));
			nextPut: (2r10000000 + (code bitShift: -12));
			nextPut: (2r10000000 + ((code bitShift: -6) bitAnd: 2r111111)); 
			nextPut: (2r10000000 + (code bitAnd: 2r111111)) ].
	self error: 'Unicode character code point outside range'
! !

Object subclass: #ZnCharacterReadStream
	instanceVariableNames: 'stream encoder peekedCharacter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnCharacterReadStream commentStamp: '<historical>' prior: 0!
I am ZnCharacterReadStream.
I wrap another binary ReadStream and use a ZnCharacerEncoder to allow Characters to be read.

I am not positionable, but I do allow a one character peek using a one character internal buffer.

Part of Zinc HTTP Components.
!

!ZnCharacterReadStream class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 14:43'!
defaultEncoder
	^ ZnUTF8Encoder new! !

!ZnCharacterReadStream class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 5/3/2012 14:40'!
on: readStream
	^ self new
		on: readStream;
		yourself! !

!ZnCharacterReadStream class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 5/3/2012 15:13'!
on: readStream encoding: encoding
	^ self new
		on: readStream;
		encoding: encoding;
		yourself! !

!ZnCharacterReadStream methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 5/3/2012 21:29'!
atEnd
	^ peekedCharacter isNil and: [ stream atEnd ]! !

!ZnCharacterReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/3/2012 14:39'!
close
	stream close! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:09'!
collectionSpecies
	^ String! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:12'!
contents
	"This is technically not correct, but it is better than nothing"
	
	^ self upToEnd! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 14:42'!
encoder
	^ encoder ifNil: [ encoder := self class defaultEncoder ]! !

!ZnCharacterReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/3/2012 14:41'!
encoder: characterEncoder
	encoder := characterEncoder! !

!ZnCharacterReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/3/2012 13:41'!
encoding: encoding
	encoder := ZnCharacterEncoder newForEncoding: encoding! !

!ZnCharacterReadStream methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:07'!
isBinary
	^ false! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:46'!
next
	^ peekedCharacter
		ifNil: [ 
			stream atEnd
				ifFalse: [ self encoder nextFromStream: stream ] ]
		ifNotNil: [ | character |
			character := peekedCharacter.
			peekedCharacter := nil.
			character ]! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:18'!
next: requestedCount 
	"Read requestedCount elements into new collection and return it,
	 it could be that less elements were available"

	^ self 
		next: requestedCount 
		into: (self collectionSpecies new: requestedCount)! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:18'!
next: requestedCount into: collection
	"Read requestedCount elements into collection,
	returning a copy if less elements are available"
	
	^ self
		next: requestedCount
		into: collection
		startingAt: 1! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:19'!
next: requestedCount into: collection startingAt: offset
	"Read requestedCount elements into collection starting at offset,
	returning a copy if less elements are available"
	
	| readCount |
	readCount := self 
		readInto: collection 
		startingAt: offset 
		count: requestedCount.
	^ requestedCount = readCount
		ifTrue: [ collection ]
		ifFalse: [ collection copyFrom: 1 to: offset + readCount - 1 ]! !

!ZnCharacterReadStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/3/2012 14:40'!
on: readStream
	stream := readStream! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:35'!
peek
	^ peekedCharacter
		ifNil: [
			stream atEnd
				ifFalse: [ 
					peekedCharacter := self encoder nextFromStream: stream ] ]
! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/3/2012 13:05'!
peekFor: object
	^ self peek = object
		ifTrue: [ 
			self next.
			true ]
		ifFalse: [ false ]! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:57'!
readInto: collection startingAt: offset count: requestedCount 
	"Read count elements and place them in collection starting at offset.
	Return the number of elements actually read.
	This is an inefficient implementation, reading characters one by one."
	
	0 to: requestedCount - 1 do: [ :count | | object |
		(object := self next) ifNil: [ ^ count ].  
		collection at: offset + count put: object ].
	^ requestedCount! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:11'!
skip: count
	count timesRepeat: [ self next ]! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:17'!
upTo: anObject 	
	^ self collectionSpecies 
		streamContents: [ :out | | element |
			[ self atEnd or: [ (element := self next) = anObject ] ] whileFalse: [ 
				out nextPut: element ] ]! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/16/2012 15:42'!
upToEnd
	^ self collectionSpecies
		streamContents: [ :collectionStream | 
			[ self atEnd ] whileFalse: [ collectionStream nextPut: self next ] ]! !

!ZnCharacterReadStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 14:41'!
wrappedStream
	^ stream! !

Object subclass: #ZnCharacterWriteStream
	instanceVariableNames: 'stream encoder'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnCharacterWriteStream commentStamp: '<historical>' prior: 0!
I am ZnCharacterWriteStream.
I wrap another binary WriteStream and use a ZnCharacerEncoder to allow Characters to be written.

Part of Zinc HTTP Components.!

!ZnCharacterWriteStream class methodsFor: 'as yet unclassified' stamp: 'SvenVanCaekenberghe 5/3/2012 15:50'!
defaultEncoder
	^ ZnUTF8Encoder new! !

!ZnCharacterWriteStream class methodsFor: 'as yet unclassified' stamp: 'SvenVanCaekenberghe 5/3/2012 15:51'!
on: writeStream
	^ self new
		on: writeStream;
		yourself! !

!ZnCharacterWriteStream class methodsFor: 'as yet unclassified' stamp: 'SvenVanCaekenberghe 5/3/2012 15:51'!
on: writeStream encoding: encoding
	^ self new
		on: writeStream;
		encoding: encoding;
		yourself! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:59'!
<< collection
	^ self nextPutAll: collection! !

!ZnCharacterWriteStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/3/2012 15:51'!
close
	stream close! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 18:47'!
encoder
	^ encoder ifNil: [ encoder := self class defaultEncoder ]! !

!ZnCharacterWriteStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/3/2012 15:51'!
encoder: characterEncoder
	encoder := characterEncoder! !

!ZnCharacterWriteStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/3/2012 15:51'!
encoding: encoding
	encoder := ZnCharacterEncoder forEncoding: encoding! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:52'!
flush
	stream flush! !

!ZnCharacterWriteStream methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:52'!
isBinary
	^ false! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:54'!
next: count putAll: collection	
	self 
		next: count 
		putAll: collection 
		startingAt: 1! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:57'!
next: count putAll: collection startingAt: offset
	"Write count characters from collection starting at offset.
	This is an inefficient implementation, writing characters one by one."
	
	0 to: count - 1 do: [ :each | 
		self nextPut: (collection at: offset + each) ]! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:56'!
nextPut: object
	self encoder 
		nextPut: object 
		toStream: stream.
	^ object! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:56'!
nextPutAll: collection
	self 
		next: collection size 
		putAll: collection 
		startingAt: 1! !

!ZnCharacterWriteStream methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/3/2012 15:51'!
on: writeStream
	stream := writeStream! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 15:52'!
space
	self nextPut: Character space! !

!ZnCharacterWriteStream methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/3/2012 18:47'!
wrappedStream
	^ stream! !

Object subclass: #ZnPercentEncoder
	instanceVariableNames: 'characterEncoder safeSet'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnPercentEncoder commentStamp: '<historical>' prior: 0!
I am ZnPercentEncoder.
I implement RFC 3986 percent encoding and decoding.

All characters that are not part of a safe set are encoded using a percent (%) followed by a two digit hexadecimal number of a byte value. Non-ASCII characters are first encoded, normally using UTF-8.

My #encode: and #decode: messages work from String to String.

My decoder additionally will accept + as an encoding for a space.

See also http://en.wikipedia.org/wiki/Percent-encoding

Part of Zinc HTTP Components
!

!ZnPercentEncoder methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/15/2012 13:53'!
characterEncoder
	"Return the character encoder that I currently use.
	If not set, I will default to using UTF-8."
	
	^ characterEncoder ifNil: [ characterEncoder := ZnUTF8Encoder new ]! !

!ZnPercentEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/15/2012 13:53'!
characterEncoder: object
	"Set the character encoding to use to object."
	
	characterEncoder := object! !

!ZnPercentEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/15/2012 13:56'!
decode: string
	"Decode a percent encoded string and return the resulting byte array.
	After percent decoding I will use my character encoder to convert byte values to characters."

	| bytes stringStream |
	stringStream := string readStream.
	bytes := ByteArray streamContents: [ :byteStream | self decode: stringStream to: byteStream ].
	^ self characterEncoder decodeBytes: bytes! !

!ZnPercentEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/16/2012 16:41'!
decode: stringStream to: byteStream
	| char |
	[ stringStream atEnd ]
		whileFalse: [ 
			(char := stringStream next) == $+
				ifTrue: [ byteStream nextPut: 32 ]
				ifFalse: [ 
					char == $%
						ifTrue: [ byteStream nextPut: (self readHexFrom: stringStream) ]
						ifFalse: [ 
							char charCode < 128
								ifTrue: [ byteStream nextPut: char charCode ]
								ifFalse: [ ZnCharacterEncodingError signal: 'ASCII character expected' ] ] ] ]! !

!ZnPercentEncoder methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/15/2012 13:56'!
encode: string
	"Encode string using percent encoding and return the resulting string.
	I will use my character encoder to convert string to bytes and then
	percent encode all byte values that are not in my safe set."

	^ String streamContents: [ :stream | self encode: string readStream to: stream ]! !

!ZnPercentEncoder methodsFor: 'converting' stamp: 'BenjaminVanRyseghem 3/1/2013 18:46'!
encode: readStream to: writeStream
	| bytes buffer byte |
	buffer := (bytes := ByteArray new: 4) writeStream.
	self safeSet; characterEncoder.
	[ readStream atEnd ]
		whileFalse: [ 
			buffer reset.
			characterEncoder nextPut: readStream next toStream: buffer.
			1 to: buffer position do: [ :index | 
				(safeSet includes: (byte := bytes at: index))
					ifTrue: [ writeStream nextPut: byte asCharacter ]
					ifFalse: [ 
						writeStream nextPut: $%.
						byte printOn: writeStream base: 16 length: 2 padded: true ] ] ]! !

!ZnPercentEncoder methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/16/2012 15:12'!
readHexFrom: stream
	| first second |
	(stream atEnd not and: [(first := stream next digitValue) between: 0 and: 15])
		ifFalse: [ ZnCharacterEncodingError signal: 'hex digit expected' ].
	(stream atEnd not and: [ (second := stream next digitValue) between: 0 and: 15 ])
		ifFalse: [ ZnCharacterEncodingError signal: 'hex digit expected' ].
	^ (first << 4) + second! !

!ZnPercentEncoder methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/15/2012 14:02'!
safeSet
	"Return the safe set of characters that I will not encode, as a byte array.
	If not set, I will default to the most commonly used safe set"
	
	^ safeSet ifNil: [ safeSet := 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.~' asByteArray ]! !

!ZnPercentEncoder methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 12/15/2012 13:58'!
safeSet: string
	"Set my safe set to be the characters in string, which I will convert to bytes"
	
	safeSet := string asByteArray ! !

Error subclass: #ZnCharacterEncodingError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Character-Encoding-Core'!

!ZnCharacterEncodingError commentStamp: '<historical>' prior: 0!
I am ZnCharacterEncodingError.
I am an Error.

I signal when something goes wrong while encoding or decoding characters.

Part of Zinc HTTP Components!
ZnBase64Encoder initialize!
ZnByteEncoder initialize!
